package cmdVM

import (
	"errors"
	"regexp"
	"strings"

	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/client/cmd"
	u "gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/client/utils"
	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/common/vm"
	nc "gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/libclient/nexusclient"
	"github.com/google/uuid"
)

type GetVMs func() ([]vm.VMNetworkSerialized, error)

func printRegexUsage(c cmd.Command) {
	for _, desc := range c.GetDesc() {
		u.PrintlnErr(desc)
	}
	u.PrintlnErr("―――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――")
	u.PrintlnErr("USAGE: ", c.GetName(), " [ID ...] [regex ...]")
	u.PrintlnErr("―――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――")
}

func printRegexUsageDetails() {
	u.PrintlnErr("The action only applies to VMs matching the specified IDs or regexes.")
	const usage string = `Any number of IDs or regexes can be specified.
The regex only matches the VM's name and is case-insensitive.
Regex examples:
"."   -> matches any VMs
"bla" -> matches any VMs containing "bla" in their name`
	u.PrintlnErr(usage)
}

// Returns a list of filtered VMs based on a regex.
// Filters are based on patterns describing IDs or regexes.
// Patterns can either be:
// - any number of "VM IDs" (UUID)
// - any number of "VM names"
// - any combination of "VM IDs" and "VM names"
// - any regular expression (it applies to the VM's name only)
// Remark: the matching is case-insensitive
// Regular expression examples:
//
//	"."   -> matches everything
//	"bla" -> matches any VM name containing "bla"
func getFilteredVMs(filteredVMsFunc GetVMs, patterns []string) ([]vm.VMNetworkSerialized, error) {
	if len(patterns) < 1 {
		return nil, errors.New("at least one ID or regex must be specified")
	}

	var ids []string
	var regexes []string

	for _, pattern := range patterns {
		_, err := uuid.Parse(pattern)
		if err != nil {
			regexes = append(regexes, pattern)
		} else {
			ids = append(ids, pattern)
		}
	}

	vms, err := filteredVMsFunc()
	if err != nil {
		return nil, err
	}

	filteredVMs := []vm.VMNetworkSerialized{}

	for _, vm := range vms {
		found := false
		for _, id := range ids {
			if id == vm.ID.String() {
				filteredVMs = append(filteredVMs, vm)
				found = true
				break
			}
		}
		if found {
			continue
		}
		for _, regex := range regexes {
			match, err := regexp.MatchString(strings.ToLower(regex), strings.ToLower(vm.Name))
			if err != nil {
				return nil, errors.New("Error matching \"" + regex + "\": " + err.Error())
			} else {
				if match {
					filteredVMs = append(filteredVMs, vm)
					break
				}
			}
		}
	}
	return filteredVMs, nil
}

// Returns a list of filtered VM credentials.
// Filters are based on patterns describing IDs or regexes.
// Patterns can either be:
// - any number of "VM IDs" (UUID)
// - any number of "VM names"
// - any combination of "VM IDs" and "VM names"
// - any regular expression (it applies to the VM's name only)
// Remark: the matching is case-insensitive
// Regular expression examples:
//
//	"."   -> matches everything
//	"bla" -> matches any VM name containing "bla"
func getFilteredAttachVMsCreds(client *nc.NexusClient, patterns []string) ([]vm.VMAttachCredentialsSerialized, error) {
	if len(patterns) < 1 {
		return nil, errors.New("at least one ID or regex must be specified")
	}

	var ids []string
	var regexes []string

	for _, pattern := range patterns {
		_, err := uuid.Parse(pattern)
		if err != nil {
			regexes = append(regexes, pattern)
		} else {
			ids = append(ids, pattern)
		}
	}

	attachVMsCreds, err := client.GetAttachVMsCreds()
	if err != nil {
		return nil, err
	}

	filteredVMAttachCreds := []vm.VMAttachCredentialsSerialized{}

	for _, vm := range attachVMsCreds {
		found := false
		for _, id := range ids {
			if id == vm.ID.String() {
				filteredVMAttachCreds = append(filteredVMAttachCreds, vm)
				found = true
				break
			}
		}
		if found {
			continue
		}
		for _, regex := range regexes {
			match, err := regexp.MatchString(strings.ToLower(regex), strings.ToLower(vm.Name))
			if err != nil {
				return nil, errors.New("Error matching \"" + regex + "\": " + err.Error())
			} else {
				if match {
					filteredVMAttachCreds = append(filteredVMAttachCreds, vm)
					break
				}
			}
		}
	}
	return filteredVMAttachCreds, nil
}
