package router

import (
	"errors"
	"net/http"
	"sync"
	"time"

	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/server/users"

	"gitedu.hesge.ch/flg_projects/nexus_vdi/nexus/common/params"

	"github.com/go-playground/validator/v10"
	"github.com/golang-jwt/jwt"
	"github.com/labstack/echo/v4"
	"github.com/labstack/echo/v4/middleware"
	"github.com/sethvargo/go-password/password"
	"golang.org/x/crypto/bcrypt"
)

// jwt.StandardClaims is added as an embedded type to provide fields like user email and expiration time.
type customClaims struct {
	Email string
	jwt.StandardClaims
}

type auth struct {
	users       *users.Users
	tokenAccess middleware.JWTConfig
}

type userRequest struct {
	email string
	ip    string
}

var loginAttempts map[userRequest]int // Number of consecutive wrong logins for the same user
var loginAttemptsMutex *sync.Mutex

var lockedUsers map[userRequest]time.Time // Locked users and timestamps when their accounts were locked
var lockedUsersMutex *sync.Mutex

var jwtSecretKey string

// Creates an auth object.
func NewAuth(u *users.Users) (*auth, error) {
	loginAttempts = make(map[userRequest]int)
	loginAttemptsMutex = new(sync.Mutex)
	lockedUsers = make(map[userRequest]time.Time)
	lockedUsersMutex = new(sync.Mutex)

	// TODO: better to store jwtSecretKey in an env. variable?
	// Generates a 64 characters long key with 20 digits, 0 symbols,
	// allowing upper and lower case letters, and allowing repeat characters.
	secret, err := password.Generate(64, 20, 0, false, true)
	if err != nil {
		return nil, errors.New("Secret generation error: " + err.Error())
	}
	jwtSecretKey = secret

	return &auth{
		users: u,
		tokenAccess: middleware.JWTConfig{
			ContextKey: "user",
			SigningKey: []byte(jwtSecretKey),
			Claims:     &customClaims{},
			ErrorHandlerWithContext: func(err error, c echo.Context) error {
				return echo.NewHTTPError(http.StatusUnauthorized, "token expired: access denied")
			},
		},
	}, nil
}

// Token used by the JWT middleware to protect access to the specified group of routes.
func (auth *auth) GetTokenAccess() middleware.JWTConfig {
	return auth.tokenAccess
}

// curl --cacert ca.pem -X POST https://localhost:1077/login -H 'Content-Type: application/json' -d '{"email": "johndoe@nexus.org", "pwd":"the_user_password"}'
// Returns a jwt token if authentication was successful:
// {"token":"<jwt token>"}
func (auth *auth) Login(c echo.Context) error {
	// Deserializes the JSON body and checks its validity.
	reqUser := new(params.Login)
	if err := decodeJson(c, &reqUser); err != nil {
		return echo.NewHTTPError(http.StatusInternalServerError, err.Error())
	}
	if err := validator.New(validator.WithRequiredStructEnabled()).Struct(reqUser); err != nil {
		log.Info("Login attempt for ", reqUser.Email, ", password validation failed, source ip=", c.RealIP())
		return echo.NewHTTPError(http.StatusBadRequest, err.Error())
	}

	// User exists?
	user, err := auth.users.GetUserByEmail(reqUser.Email)
	if err != nil {
		log.Info("Login attempt for non-existing user ", reqUser.Email, ", source ip=", c.RealIP())
		return echo.NewHTTPError(http.StatusUnauthorized, err.Error())
	}

	// A user request is the combination of user email and where the request comes from (ip address)
	userReq := userRequest{email: user.Email, ip: c.RealIP()}

	// Is user locked?
	lockedUsersMutex.Lock()
	lockedTime, found := lockedUsers[userReq]
	if found {
		lockedUsersMutex.Unlock()
		// If locked time has elapsed, unlock the user.
		t := time.Now()
		elapsedTime := t.Sub(lockedTime)
		if elapsedTime.Hours() > conf.Auth.LockedTimeInHours {
			if err := UnlockUser(c, user.Email); err != nil {
				return echo.NewHTTPError(http.StatusInternalServerError, err.Error())
			}
		} else {
			log.Info("User ", user.Email, " is locked, source ip=", c.RealIP())
			return echo.NewHTTPError(http.StatusUnauthorized, "User is locked")
		}
	} else {
		lockedUsersMutex.Unlock()
	}

	// Checks the user password is correct.
	// If a user enters the wrong password too many times, the user will be locked for a number of hours.
	err = bcrypt.CompareHashAndPassword([]byte(user.Pwd), []byte(reqUser.Pwd))

	// If err == nil, passwords match!
	if err == nil {
		loginAttemptsMutex.Lock()
		loginAttempts[userReq] = 0
		loginAttemptsMutex.Unlock()
	} else {
		// Stores the number of attempted logins.
		loginAttemptsMutex.Lock()
		count, found := loginAttempts[userReq]
		if found {
			count = count + 1
		} else {
			count = 1
		}
		loginAttempts[userReq] = count
		loginAttemptsMutex.Unlock()
		log.Info("Login attempt ", count, " for user ", user.Email, ", source ip=", c.RealIP())
		if count > conf.Auth.MaxLoginAttempts {
			// Locks user
			lockedUsersMutex.Lock()
			lockedUsers[userReq] = time.Now()
			lockedUsersMutex.Unlock()
			log.Info("User ", user.Email, " is locked: too many wrong password attempts, source ip=", c.RealIP())
			return echo.NewHTTPError(http.StatusUnauthorized, "User is locked")
		}

		// If passwords don't match, return a 401 status.
		return echo.NewHTTPError(http.StatusUnauthorized, "Invalid password")
	}

	tok, err := createToken(user.Email)
	if err != nil {
		return echo.NewHTTPError(http.StatusUnauthorized, err.Error())
	}

	tokenSerialized := params.Token{Token: tok}

	log.Info("Successful login for ", user.Email, ", source ip=", c.RealIP())
	return c.JSON(http.StatusOK, tokenSerialized)
}

// Creates a JWT token and claims with the expiration time for the given user.
func createToken(userEmail string) (string, error) {
	expirationTime := time.Now().Add(time.Duration(conf.Auth.AccessTokenExpirationTimeInHours) * time.Hour)

	// Creates the claims
	claims := &customClaims{
		Email: userEmail,
		StandardClaims: jwt.StandardClaims{
			// JWT expresses expiration time in Unix milliseconds
			ExpiresAt: expirationTime.Unix(),
		},
	}

	// Declares the token with the algorithm used for signing, and the claims.
	token := jwt.NewWithClaims(jwt.SigningMethodHS256, claims)

	// Signs the token and gets a string version of the token.
	secret := []byte(jwtSecretKey)
	tokenString, err := token.SignedString(secret)
	if err != nil {
		return "", err
	}

	return tokenString, nil
}

// Returns user email from the Echo context.
// If an error occured, returns the error.
func GetUserFromContext(c echo.Context) (string, error) {
	if c.Get("user") == nil {
		return "", errors.New("user not authenticated")
	}

	user := c.Get("user").(*jwt.Token)
	claims := user.Claims.(*customClaims)

	return claims.Email, nil
}

func UnlockUser(c echo.Context, emailToUnlock string) error {
	userEmail, err := GetUserFromContext(c)
	if err != nil {
		return err
	}

	lockedUsersMutex.Lock()
	defer lockedUsersMutex.Unlock()

	// Remove all entries matching the user email to unlock
	userFound := false
	for _userReq := range lockedUsers {
		if _userReq.email == emailToUnlock {
			delete(lockedUsers, _userReq)

			loginAttemptsMutex.Lock()
			delete(loginAttempts, _userReq) // reset the number of attempts
			loginAttemptsMutex.Unlock()

			log.Info("User ", userEmail, " unlocked user {", _userReq.email, ",", _userReq.ip, "}, source ip=", c.RealIP())
			userFound = true
		}
	}

	if userFound {
		return nil
	} else {
		return errors.New("User not locked")
	}
}

/*
// Example of middleware
// Such a middleware could be used to handle a logout feature.
// We would keep a map of invalid tokens (the ones that logged out), and deny access to such tokens.
// Question: Entries in the map would have to be cleared at times to prevent it from growing too much.
//           This could be done when the token has been in the map for longer than its expiration time.

usersGroup.Use(auth.OnlyJoe)  // apply the middleware (to be added in router.go)

func OnlyJoe(next echo.HandlerFunc) echo.HandlerFunc {
    return func(c echo.Context) error {
        // Gets user token from context.
        if c.Get("user") == nil {
            return echo.NewHTTPError(http.StatusUnauthorized, "invalid user token\n")
        }
        user := c.Get("user").(*jwt.Token)
        claims := user.Claims.(*customClaims)
        username := claims.Username
        if username != "joe" {
            return echo.ErrUnauthorized
        }
        return next(c)
    }
}
*/
